import std.traits : fullyQualifiedName;

void main() @safe
{
  import std.stdio : writeln;

  enum ubyte[5][5][] boards = [
      [
        [67, 64, 11, 28, 16],
        [63, 26, 20, 15, 10],
        [68, 44, 0_, 53, 70],
        [22, 56, 38, 51, 9_],
        [47, 33, 17, 39, 59],
      ],
      [
        [51, 24, 53, 70, 62],
        [54, 44, 57, 72, 35],
        [32, 5_, 0_, 20, 38],
        [36, 4_, 73, 29, 69],
        [63, 42, 7_, 8_, 58],
      ],
    ];
  enum ubyte[] bingos = [
      0, 72, 3, 8, 59, 66, 61, 58, 23, 14, 16, 42, 10, 17, 2, 48, 44, 26, 70, 21,
      31, 19, 9
    ];
  enum ubyte[] bingos2 = [
      72, 101, 108, 108, 111, 44, 32, 119, 111, 114, 108, 100, 33
    ];
  enum ubyte[] bingos3 = [0, 51, 54, 32, 36, 63];
  enum ubyte[] bingos4 = [0, 51, 44, 29, 58];
  string hello()
  {
    return bingos2;
  }

  assert(ones * ones * (ones * ones + ones * ones) - ones * ones == sentinel + 3);

  entry(boards, bingos).writeln;
}

struct Card
{
  ubyte[5][5] inner_;
  alias inner_ this;

  mixin(overload!Card("*", q{
    res[i][j] += this[i][k] * rhs[k][j];
  }, ["i", "j", "k"]));

  mixin(overload!Strip("*", q{
    res[i] += this[i][j] * rhs[j];
  }, ["i", "j"]));

  mixin(overload!Card("~", q{
    res[i][j] += this[i][j] * rhs[i][j];
  }, ["i", "j"]));

  mixin(overload!Card("in", q{
    res[i][j] = this[i][j] == rhs;
  }, ["i", "j"], "ubyte"));

  void opOpAssign(string op, T)(const T rhs)
  {
    mixin("this = this " ~ op ~ "rhs;");
  }

  mixin(overload!Card("*", q{
    res[i][j] = this[j][i];
  }, ["i", "j"], ""));

  mixin(overload!Card("~", q{
    res[i][j] = !this[i][j];
  }, ["i", "j"], ""));

  pure Strip diag() const nothrow @safe @nogc @property
  {
    Strip res;
    foreach (i; 0 .. 5)
      res[i] += Strip(i) * this * Strip(i);
    return res;
  }
}

struct Strip
{
  ubyte[5] inner_;
  alias inner_ this;

  pure nothrow this(const size_t one_hot_index) const @safe @nogc
  {
    this[one_hot_index] = 1;
  }

  mixin(overload!Strip("*", q{
    res[j] += this[i] * rhs[i][j];
  }, ["i", "j"], "Card"));

  mixin(overload!ubyte("*", q{
    res += this[i] * rhs[i];
  }, ["i"], "Strip"));

  mixin(overload!ubyte("~", q{
    res *= this[i] * rhs[i];
  }, ["i"], "Strip", " = memey.diag * ymeme.right"));

  mixin(overload!Strip("*", q{
    res[i] = cast(bool) this[i];
  }, ["i"], ""));

  mixin(overload!ubyte("~", q{
    res[i] = !this[i];
  }, ["i"], ""));
}

enum ymeme = Card([
    [0, 0, 'ÿ', 0, 0],
    [0, 0, 0, 'm', 0],
    [0, 0, 0, 0, 'e'],
    ['m', 0, 0, 0, 0],
    [0, 'e', 0, 0, 0],
  ]);

enum memey = Card([
    [0, 'e', 0, 0, 0],
    ['m', 0, 0, 0, 0],
    [0, 0, 0, 0, 'ÿ'],
    [0, 0, 0, 'e', 0],
    [0, 0, 'm', 0, 0],
  ]);

enum involutory_perm = ymeme * memey;

enum identity = involutory_perm * involutory_perm;

enum sentinel = 42;

enum ones = identity.diag;

pure R left(R = Strip, I)(const I a) @property
{
  return ones * a;
}

pure R right(R = Strip, I)(const I a) @property
{
  return a * ones;
}

string overload(T)(string op, string code, string[] vars, string paramType = fullyQualifiedName!T, string init = "")
{
  import std.format;

  string nameInfix = paramType == "" ? "U" : "Bi";
  string param = paramType == "" ? "" : "const " ~ paramType ~ " rhs";
  string loops = q{%-(foreach (%1$s; 0 .. 5) %|%)}.format(vars);
  return q{
    import std.stdio;
    pure %1$s op%5$snary(string op : "%2$s")(%6$s) const
    {
      %1$s res%7$s;
      %3$s%4$s
      return res;
    }
  }.format(fullyQualifiedName!T, op, loops, code, nameInfix, param, init);
}

/**
 * Decides the winner of a bingo game.
 *
 * Gramma Whetherwhacks stared at her bingo card. Nothing crossed, not a digit!
 * It was as if an occult hand had reached into the hat of numbers while no one was looking and stolen her chance of success.
 * From across the hall, “Bingo!„
 * “Curse you, Hadamard!„ she shouted at her rival.
 * Hadamard would later speak of feeling like a white-hot hand pressing on his shoulder, and die of a heart attack before he could collect the prize.
 *
 * Example:
 * --------------------------------------
 * size_t winner = entry(boards, bingos);
 * --------------------------------------
 *
 * Params:
 *   cards = the bingo cards players have
 *   calls = the numbers drawn
 *
 * Returns:
 *   Index of winner
 */
public size_t entry(const ubyte[5][5][] cards, const ubyte[] calls) pure nothrow @safe @nogc
{
  size_t res;
  foreach (i, squares; cards)
  {
    Card card = Card(squares);
    Card matrix = ~(*ymeme ~ ymeme);
    foreach (call; calls)
      matrix ~= ~(card in call);
    res += matrix.diag.right!ubyte * (involutory_perm * matrix)
      .diag.left!ubyte * (*matrix.right ~ *matrix.left) ? 0 : i;
  }
  return res;
}
